/* spfmilter - SPF mail filter module
**
** Copyright  2004 by Jef Poskanzer <jef@mail.acme.com>.
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**	notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**	notice, this list of conditions and the following disclaimer in the
**	documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
**
** For commentary on this license please see http://www.acme.com/license.html
*/


#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <sysexits.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <getopt.h>
#include <pthread.h>
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include <libmilter/mfapi.h>

#include "version.h"
#include "match.h"
#include "iptab.h"


/* Defines. */

#define HEADER_NAME "Received-SPF"

#define SPFMILTER_RESULT_PASS 0
#define SPFMILTER_RESULT_FAIL 1
#define SPFMILTER_RESULT_SOFTFAIL 2
#define SPFMILTER_RESULT_NEUTRAL 3
#define SPFMILTER_RESULT_UNKNOWN 4
#define SPFMILTER_RESULT_UNMECH 5
#define SPFMILTER_RESULT_INVALID 6
#define SPFMILTER_RESULT_TEMPERROR 7
#define SPFMILTER_RESULT_ERROR 8
#define SPFMILTER_RESULT_NONE 9

#define SPFMILTER_ACTION_UNKNOWN 0
#define SPFMILTER_ACTION_REJECT 1
#define SPFMILTER_ACTION_MARK 2
#define SPFMILTER_ACTION_TEMPFAIL 3


/* Structs. */


/* Fallback list. */
struct lib_fallback_s;
typedef struct lib_fallback_s lib_fallback_t;
typedef struct {
    char* pattern;
    char* str;
    lib_fallback_t* lib_fallback;
    } fallback_t;
static fallback_t* fallbacks;
static int max_fallbacks, num_fallbacks;


/* Per-connection data structure. */
#define DEL_HEADER_MAX 10
struct lib_data_s;
typedef struct lib_data_s lib_data_t;
typedef struct {
    int whitelisted;
    int authenticated;
    char* ip_str;
    char* helo;
    char* mail_from;
    char* use_mail_from;
    lib_data_t* lib_data;
    int result;
    int action;
    int n_my_headers;
    int n_del_headers;
    int del_header_list[DEL_HEADER_MAX];
    char buf1[1000];
    char buf2[2000];
    } connection_data_t;

/* A word about buf1 and buf2, above.  These are scratch buffers for
** temporary use by any routine that needs them.  Normally we'd just
** declare them locally, on the stack.  The problem is that we are
** running in a thread, each thread has a separate stack of fixed size,
** we don't know what that size is, and if we exceed it we dump core.
** So we really want to avoid using up the stack with large local variable
** frames; instead, we use the per-connection data object which is
** allocated from the heap.  It makes the connection objects fairly
** large, but not unreasonably so.
**
** We do have to be a little careful using the buffers, so that two
** routines in the same call stack don't try to make use of them
** simultaneously.  Indeed, there's one case where we need two buffers
** at the same time so that's why we have two here.
*/


/* Forwards. */

static void usage( void );
static void init_fallback( const char* fallback_filename, const char* guess_str );
static void add_fallback( const char* pattern, const char* str );
static void term_fallback( void );
static void init_whitelist( const char* whitelist_filename );
static void term_whitelist( void );
static void trim( char* str );
static void init_uid( const char* user );
static void init_socket( char* sockpath );
static void term_socket( char* sockpath );
static void init_daemon( void );
static void init_pidfile( const char* pidfilename );
static void term_pidfile( const char* pidfilename );

/* The milter calback routines.  Signatures for these are fixed
** by the milter API.
*/
static sfsistat spf_connect( SMFICTX* ctx, char* connhost, _SOCK_ADDR* connaddr );
static sfsistat spf_helo( SMFICTX* ctx, char* helohost );
static sfsistat spf_envfrom( SMFICTX* ctx, char** fromargs );
static sfsistat spf_envrcpt( SMFICTX* ctx, char** rcptargs );
static sfsistat spf_header( SMFICTX* ctx, char* name, char* value );
static sfsistat spf_eoh( SMFICTX* ctx );
static sfsistat spf_eom( SMFICTX* ctx );
static sfsistat spf_abort( SMFICTX* ctx );
static sfsistat spf_close( SMFICTX* ctx );

static char* trim_address( char* address );
static sfsistat handle_result( SMFICTX* ctx );
static void escape_percents( const char* in, char* out, int out_size );
static int receiver_is_me( const char* value );
static void build_header( connection_data_t* cd, char* header, size_t header_size, const char* reason );
static connection_data_t* init_connection_data( void );
static int init_message_data( connection_data_t* cd );
static void term_message_data( connection_data_t* cd );
static void term_connection_data( connection_data_t* cd );
static char* result_str( int result );
static void reason_str( connection_data_t* cd, char* str, size_t size );

/* Library interface routines.  These are implemented in parallel for
** both libspf and libspf2.
*/
static int lib_init( void );
static const char* lib_version( void );
static lib_fallback_t* lib_init_fallback( const char* str );
static int lib_set_local_hostname( lib_data_t* ld );
static lib_data_t* lib_init_connection_data( void );
static int lib_init_message_data( lib_data_t* ld );
static int lib_set_ipv4( lib_data_t* ld, struct in_addr ipv4_addr, char* ipv4_str );
static int lib_set_ipv6( lib_data_t* ld, struct in6_addr ipv6_addr, char* ipv6_str );
static int lib_set_helo_hostname( lib_data_t* ld, char* helo_hostname );
static int lib_set_from( lib_data_t* ld, const char* from );
static int lib_do_check( lib_data_t* ld, const char* from );
static int lib_do_check_recipient( lib_data_t* ld, const char* to );
static int lib_do_check_final( lib_data_t* ld );
static int lib_get_result( lib_data_t* ld );
static const char* lib_get_explanation( lib_data_t* ld );
static const char* lib_get_error( lib_data_t* ld );
static void lib_term_message_data( lib_data_t* ld );
static void lib_term_connection_data( lib_data_t* ld );
static void lib_term_fallback( lib_fallback_t* lf );
static void lib_term( void );


/* Globals. */

static char* argv0;

static char* localpolicy_str;
static int trustedforwarders;
static iptab whitelist;
int recipientmx;
static char* explanation_str;
static int markonly;
static int debug;
static char* local_hostname;
static int local_hostname_len;

static int header_name_len;

static struct smfiDesc smfilter = {
    "SPF",				/* filter name */
    SMFI_VERSION,			/* version code -- do not change */
    SMFIF_CHGHDRS|SMFIF_ADDHDRS,	/* flags */
    spf_connect,			/* connection info filter */
    spf_helo,				/* SMTP HELO command filter */
    spf_envfrom,			/* envelope sender filter */
    spf_envrcpt,			/* envelope recipient filter */
    spf_header,				/* header filter */
    spf_eoh,				/* end of header */
    NULL,				/* body block filter */
    spf_eom,				/* end of message */
    spf_abort,				/* message aborted */
    spf_close,				/* connection cleanup */
    NULL,				/* unrecognized / unimplemented cmd */
    NULL,				/* DATA command filter  */
    NULL				/* negotiation  */
    };

#ifdef HAVE_GETOPT_LONG
static const struct option	longopts[] = {
    { "localpolicy",		required_argument,	NULL,	'l', },
    { "trustedforwarders",	no_argument,		NULL,	't', },
    { "guess",			required_argument,	NULL,	'g', },
    { "fallback",		required_argument,	NULL,	'f', },
    { "whitelist",		required_argument,	NULL,	'w', },
    { "recipientmx",		no_argument,		NULL,	'r', },
    { "explanation",		required_argument,	NULL,	'e', },
    { "markonly",		no_argument,		NULL,	'm', },
    { "user",			required_argument,	NULL,	'u', },
    { "pidfile",		required_argument,	NULL,	'p', },
    { "nodaemon",		no_argument,		NULL,	'X', },
    { "help",			no_argument,		NULL,	'h', },
    { "debug",			optional_argument,	NULL,	'd', },
    { 0, 0, 0, 0 },
};
#define DOC_LONGOPT(l, v, t, p1) \
    do { \
	(void) fprintf( stderr, "    --%s%c%s%*s" t "\n", l, (v ? '=' : ' '),  (v ? v : ""), p1, "" ); \
    } while( 0 )
#else /* HAVE_GETOPT_LONG */
#define DOC_LONGOPT(l, v, t, p1) do { } while( 0 )
#endif /* HAVE_GETOPT_LONG */
static const char* shortopts = "l:tg:f:w:re:mu:p:Xhd::";

#define DOC_OPT(s, l, v, t, p0, p1) \
    do { \
	(void) fprintf( stderr, "    -%c%c%s%*s" t "\n", s, (v ? ' ' : ' '), (v ? v : ""), p0, "" ); \
	DOC_LONGOPT(l, v, t, p1); \
    } while( 0 )


int
main( int argc, char** argv )
    {
    char* guess_str;
    char* fallback_filename;
    char* whitelist_filename;
    char* user;
    char* pidfilename;
    int nodaemon;
    char* sockpath;
#ifdef HAVE_GETOPT_LONG
    int idx;
#endif /* HAVE_GETOPT_LONG */
    char c;

    localpolicy_str = (char*) 0;
    trustedforwarders = 0;
    guess_str = (char*) 0;
    fallback_filename = (char*) 0;
    whitelist_filename = (char*) 0;
    recipientmx = 0;
    explanation_str = (char*) 0;
    markonly = 0;
    user = (char*) 0;
    pidfilename = (char*) 0;
    nodaemon = 0;
    debug = 0;

    /* Figure out the program's name. */
    argv0 = strrchr( argv[0], '/' );
    if ( argv0 != (char*) 0 )
	++argv0;
    else
	argv0 = argv[0];

    openlog( argv0, LOG_PERROR, LOG_MAIL );

    header_name_len = strlen( HEADER_NAME );

    /* Parse args. */
    while ( ( c =
#ifdef HAVE_GETOPT_LONG
		  getopt_long( argc, argv, shortopts, longopts, &idx )
#else /* HAVE_GETOPT_LONG */
		  getopt( argc, argv, shortopts )
#endif /* HAVE_GETOPT_LONG */
			) != -1 )
	{
	switch (c)
	    {
	    case 'l':
		localpolicy_str = optarg;
		break;
	    case 't':
		trustedforwarders = 1;
		break;
	    case 'g':
		guess_str = optarg;
		break;
	    case 'f':
		fallback_filename = optarg;
		break;
	    case 'w':
		whitelist_filename = optarg;
		break;
	    case 'r':
		recipientmx = 1;
		break;
	    case 'e':
		explanation_str = optarg;
		break;
	    case 'm':
		markonly = 1;
		break;
	    case 'u':
		user = optarg;
		break;
	    case 'p':
		pidfilename = optarg;
		break;
	    case 'X':
		nodaemon = 1;
		break;
	    case 'h':
	    case '?':
		usage();
		exit( EX_USAGE );
	    case 'd':
		if ( optarg )
		    debug = atoi( optarg );
		else
		    debug = 1;
		break;
	    default:
		(void) fprintf( stderr, "unrecognised option '%c'\n", c );
		exit( EX_USAGE );
	    }
	}
    argv += optind;
    argc -= optind;

    if ( argc != 1 )
	{
	usage();
	exit( EX_USAGE );
	}
    sockpath = argv[0];

    local_hostname = (char*) 0;
    if ( ! lib_init() )
	{
	syslog( LOG_ERR, "lib_init() failed" );
	exit( EX_OSERR );
	}
    init_fallback( fallback_filename, guess_str );
    init_whitelist( whitelist_filename );
    init_socket( sockpath );
    init_pidfile( pidfilename );
    init_uid( user );
    if ( ! nodaemon )
	init_daemon();

    syslog( LOG_NOTICE, "%s %s with %s starting", SPFMILTER_PROGRAM, SPFMILTER_VERSION, lib_version() );
    if ( smfi_main() == MI_FAILURE )
	{
	syslog( LOG_ERR, "smfi_main() failed" );
	exit( EX_OSERR );
	}
    syslog( LOG_NOTICE, "%s %s with %s terminating", SPFMILTER_PROGRAM, SPFMILTER_VERSION, lib_version() );

    term_pidfile( pidfilename );
    term_socket( sockpath );
    term_whitelist();
    term_fallback();
    lib_term();
    if ( local_hostname != (char*) 0 )
	free( (void*) local_hostname );

    closelog();

    return 0;
    }


static void
usage( void )
    {
    (void) fprintf( stderr, "usage: %s [options] /path/to/socket\n", argv0 );
    DOC_OPT( 'l', "localpolicy", "<spf-mechanism>", "Add local SPF policy mechanisms.", 5, 4 );
    DOC_OPT( 't', "trustedforwarders", (char*) 0, "Use the trusted-forwarder.org whitelist.", 20, 14 );
    DOC_OPT( 'g', "guess", "<spf-mechanisms>", "Add guess SPF policy mechanisms.", 4, 4 );
    DOC_OPT( 'f', "fallback", "<filename>", "A file of fallback SPF mechanisms.", 10, 14 );
    DOC_OPT( 'w', "whitelist", "<filename>", "A file of IP addresses to always accept mail from.", 10, 14 );
    DOC_OPT( 'r', "recipientmx", (char*) 0, "Check the MX records of the message's reqipients.", 20, 14 );
    DOC_OPT( 'e', "explanation", "<exp>", "Change the message returned in mail bounces.", 15, 14 );
    DOC_OPT( 'm', "markonly", (char*) 0, "Mark email instead of rejecting it.", 20, 22 );
    DOC_OPT( 'u', "user", "<user|uid>", "Run as specified user or UID.", 10, 16 );
    DOC_OPT( 'p', "pidfile", "<filename>", "Write the process i.d. to the specified file.", 10, 16 );
    DOC_OPT( 'X', "nodaemon", (char*) 0, "Do not fork into the background.", 20, 26 );
    DOC_OPT( 'h', "help", (char*) 0, "Show this help.", 20, 26 );
    DOC_OPT( 'd', "debug", "[<int>]", "Enable debugging to syslog.", 13, 18 );
    }


static void
init_fallback( const char* fallback_filename, const char* guess_str )
    {
    FILE* fp;
    char line[1000];
    char* cp;

    max_fallbacks = num_fallbacks = 0;
    fallbacks = (fallback_t*) 0;

    if ( fallback_filename != (char*) 0 )
	{
	fp = fopen( fallback_filename, "r" );
	if ( fp == (FILE*) 0 )
	    {
	    syslog( LOG_ERR, "fopen '%s' - %m", fallback_filename );
	    exit( EX_OSERR );
	    }
	while ( fgets( line, sizeof(line), fp ) != (char*) 0 )
	    {
	    trim( line );
	    if ( line[0] == '\0' )
		continue;
	    cp = line;
	    cp += strcspn( cp, " \t" );
	    *cp = '\0';
	    ++cp;
	    cp += strspn( cp, " \t" );
	    add_fallback( line, cp );
	    }
	(void) fclose( fp );
	}

    if ( guess_str != (char*) 0 )
	add_fallback( "*", guess_str );
    }


static void
add_fallback( const char* pattern, const char* str )
    {
    if ( num_fallbacks >= max_fallbacks )
	{
	if ( max_fallbacks == 0 )
	    {
	    max_fallbacks = 50;		/* arbitrary */
	    fallbacks = (fallback_t*) malloc( max_fallbacks * sizeof(fallback_t) );
	    }
	else
	    {
	    max_fallbacks *= 2;
	    fallbacks = (fallback_t*) realloc( (void*) fallbacks, max_fallbacks * sizeof(fallback_t) );
	    }
	if ( fallbacks == (fallback_t*) 0 )
	    {
	    syslog( LOG_ERR, "out of memory enlarging fallbacks array" );
	    exit( EX_OSERR );
	    }
	}
    fallbacks[num_fallbacks].pattern = strdup( pattern );
    fallbacks[num_fallbacks].str = strdup( str );
    if ( fallbacks[num_fallbacks].pattern == (char*) 0 ||
         fallbacks[num_fallbacks].str == (char*) 0 )
	{
	syslog( LOG_ERR, "out of memory saving a fallback entry" );
	exit( EX_OSERR );
	}

    fallbacks[num_fallbacks].lib_fallback = lib_init_fallback( fallbacks[num_fallbacks].str );
    if ( fallbacks[num_fallbacks].lib_fallback == (lib_fallback_t*) 0 )
	{
	syslog( LOG_ERR, "error compiling fallback mechanism '%s'", fallbacks[num_fallbacks].str );
	exit( EX_OSERR );
	}

    ++num_fallbacks;
    }


static void
term_fallback( void )
    {
    if ( fallbacks != (fallback_t*) 0 )
	{
	int i;

	for ( i = 0; i < num_fallbacks; ++i )
	    {
	    free( (void*) fallbacks[i].pattern );
	    free( (void*) fallbacks[i].str );
	    lib_term_fallback( fallbacks[i].lib_fallback );
	    }
	max_fallbacks = num_fallbacks = 0;
	free( (void*) fallbacks );
	fallbacks = (fallback_t*) 0;
	}
    }


static void
init_whitelist( const char* whitelist_filename )
    {
    FILE* fp;
    char line[500];
    ipaddress ipa;

    if ( whitelist_filename != (char*) 0 )
	{
	whitelist = iptab_new();
	if ( whitelist == (iptab) 0 )
	    {
	    syslog( LOG_ERR, "whitelist create failed" );
	    exit( EX_OSERR );
	    }
	fp = fopen( whitelist_filename, "r" );
	if ( fp == (FILE*) 0 )
	    {
	    syslog( LOG_ERR, "fopen '%s' - %m", whitelist_filename );
	    exit( EX_OSERR );
	    }
	while ( fgets( line, sizeof(line), fp ) != (char*) 0 )
	    {
	    trim( line );
	    if ( line[0] == '\0' )
		continue;
	    if ( iptab_parse_address( line, &ipa ) )
		(void) iptab_add( whitelist, &ipa );
	    else
		syslog( LOG_WARNING, "unparsable IP address - '%s'", line );
	    }
	(void) fclose( fp );
	}
    else
	whitelist = (iptab) 0;
    }


static void
term_whitelist( void )
    {
    if ( whitelist != (iptab) 0 )
	iptab_delete( whitelist );
    }


static void
trim( char* str )
    {
    char* cp;
    int len;

    cp = strchr( str, '#' );
    if ( cp != (char*) 0 )
	*cp = '\0';
    len = strlen( str );
    while ( str[len-1] == '\n' || str[len-1] == '\r' || str[len-1] == ' ' || str[len-1] == '\t' )
	{
	--len;
	str[len] = '\0';
	}
    }


static void
init_uid( const char* user )
    {
    struct passwd* pwd;
    char* ep;
    int uid;

    if ( getuid() == 0 )
	{
	/* If we're root, the very first thing we do is become another user. */
	if ( user == (char*) 0 )
	    syslog( LOG_WARNING, "warning: started as root but no --user flag specified" );
	else
	    {
	    /* Is it a number? */
	    uid = strtol( user, &ep, 0 );
	    if ( *ep == '\0' )
		pwd = getpwuid( uid );
	    else
		pwd = getpwnam( user );
	    if ( pwd == (struct passwd*) 0 )
		{
		syslog( LOG_ERR, "unknown user: '%s'", user );
		exit( EX_OSERR );
		}
	    /* Set aux groups to null. */
	    if ( setgroups( 0, (gid_t*) 0 ) < 0 )
		{
		syslog( LOG_ERR, "setgroups - %m" );
		exit( EX_OSERR );
		}
	    /* Set primary group. */
	    if ( setgid( pwd->pw_gid ) < 0 )
		{
		syslog( LOG_ERR, "setgid - %m" );
		exit( EX_OSERR );
		}

	    /* Try setting aux groups correctly - not critical if this fails. */
	    if ( initgroups( user, pwd->pw_gid ) < 0 )
		syslog( LOG_WARNING, "initgroups - %m" );
	    /* Set uid. */
	    if ( setuid( pwd->pw_uid ) < 0 )
		{
		syslog( LOG_ERR, "setuid - %m" );
		exit( EX_OSERR );
		}
	    }
	}
    else
	{
	/* If we're not root but got a -user flag anyway, that's an error. */
	if ( user != (char*) 0 )
	    {
	    syslog( LOG_ERR, "can't switch users if not started as root" );
	    exit( EX_OSERR );
	    }
	}
    }


static void
init_socket( char* sockpath )
    {
    /* Harden our umask so that the new socket gets created securely. */
    umask( 0077 );

    /* Initialize milter stuff. */
    if ( smfi_register( smfilter ) == MI_FAILURE )
	{
	syslog( LOG_ERR, "smfi_register() failed" );
	exit( EX_OSERR );
	}
    smfi_setconn( (char*) sockpath );
    if ( smfi_opensocket( true ) == MI_FAILURE )
	{
	syslog( LOG_ERR, "smfi_opensocket() failed" );
	exit( EX_OSERR );
	}
    }


static void
term_socket( char* sockpath )
    {
    }


static void
init_daemon( void )
    {
    /* Daemonize. */
#ifdef HAVE_DAEMON
    if ( daemon( 0, 0 ) < 0)
	{
	syslog( LOG_ERR, "daemon - %m" );
	exit( EX_OSERR );
	}
#else /* HAVE_DAEMON */
    switch ( fork() )
	{
	case 0:
	    break;  
	case -1:
	    syslog( LOG_ERR, "fork - %m" );
	    exit( EX_OSERR );
	default:
	    exit( EX_OK );
	}
#ifdef HAVE_SETSID
    setsid();
#endif /* HAVE_SETSID */
#endif /* HAVE_DAEMON */
    }


static void
init_pidfile( const char* pidfilename )
    {
    if ( pidfilename != (char*) 0 )
	{
	FILE* fp;

	fp = fopen( pidfilename, "w" );
	if ( fp == (FILE*) 0 )
	    syslog( LOG_ERR, "unable to write PID file - %m" );
	else
	    {
	    (void) fprintf( fp, "%ld\n", (long) getpid() );
	    (void) fclose( fp );
	    }
	}
    }


static void
term_pidfile( const char* pidfilename )
    {
    if ( pidfilename != (char*) 0 )
	(void) unlink( pidfilename );
    }


/* Milter routines. */


/* spf_connect - handle the initial TCP connection
**
** Called at the start of a connection.  Any per-connection data should
** be initialized here.
**
** connhost: The hostname of the client, based on a reverse lookup.
** connaddr: The client's IP address, based on getpeername().
*/
static sfsistat
spf_connect( SMFICTX* ctx, char* connhost, _SOCK_ADDR* connaddr )
    {
    connection_data_t* cd;
    const char* r;

    if ( connaddr == (_SOCK_ADDR*) 0 )
	return SMFIS_ACCEPT;	/* can't deal with it */

    /* Create & initialize the connection_data object. */
    cd = init_connection_data();
    if ( cd == (connection_data_t*) 0 )
	{
	syslog( LOG_ERR, "couldn't allocate connection_data" );
	return SMFIS_ACCEPT;
	}
    if ( smfi_setpriv( ctx, (void*) cd ) != MI_SUCCESS )
	{
	syslog( LOG_ERR, "smfi_setpriv() failed" );
	term_connection_data( cd );
	return SMFIS_ACCEPT;
	}

    /* The first time through, we figure out the local hostname. */
    if ( local_hostname == (char*) 0 )
	{
	char* jmac;
	char lh[200];

	jmac = smfi_getsymval( ctx, "{j}" );
	if ( jmac != (char*) 0 )
	    local_hostname = strdup( jmac );
	else
	    {
	    if ( gethostname( lh, sizeof(lh) ) == 0 )
		local_hostname = strdup( lh );
	    else
		{
		syslog( LOG_ERR, "couldn't get local hostname" );
		return SMFIS_ACCEPT;
		}
	    }
	if ( local_hostname == (char*) 0 )
	    {
	    syslog( LOG_ERR, "couldn't strdup local hostname" );
	    return SMFIS_ACCEPT;
	    }
	local_hostname_len = strlen( local_hostname );
	if ( ! lib_set_local_hostname( cd->lib_data ) )
	    {
	    syslog( LOG_ERR, "lib_set_local_hostname() failed - %s", lib_get_error( cd->lib_data ) );
	    return SMFIS_ACCEPT;
	    }
	}

    /* Set the IP address. */
    if ( connaddr->sa_family == AF_INET )
	{
	struct sockaddr_in* sa_in;

	sa_in = (struct sockaddr_in*) ( (void*) connaddr );	/* extra cast to avoid alignment warning */
	if ( whitelist != (iptab) 0 )
	    {
	    unsigned char* uchar_addr = (unsigned char*) &sa_in->sin_addr.s_addr;
	    ipaddress ipa;
	    char str[100];

	    ipa.octets[0] = ipa.octets[1] = ipa.octets[2] = ipa.octets[3] = ipa.octets[4] = ipa.octets[5] = ipa.octets[6] = ipa.octets[7] = ipa.octets[8] = ipa.octets[9] = 0;
	    ipa.octets[10] = ipa.octets[11] = 0xff;
	    ipa.octets[12] = uchar_addr[0];
	    ipa.octets[13] = uchar_addr[1];
	    ipa.octets[14] = uchar_addr[2];
	    ipa.octets[15] = uchar_addr[3];
	    ipa.prefixlen = 128;
	    if ( iptab_check( whitelist, &ipa ) )
		{
		syslog( LOG_INFO, "whitelist \"%s\" [%s]", connhost, iptab_format_address( &ipa, str, sizeof(str) ) );
		cd->whitelisted = 1;
		}
	    }
	r = inet_ntop( AF_INET, &sa_in->sin_addr, cd->buf1, sizeof(cd->buf1) );
	if ( r == (char*) 0 )
	    {
	    syslog( LOG_ERR, "couldn't convert IPv4 address to a string - %m" );
	    return SMFIS_ACCEPT;
	    }
	if ( ! lib_set_ipv4( cd->lib_data, sa_in->sin_addr, cd->buf1 ) )
	    {
	    syslog( LOG_ERR, "lib_set_ipv4() failed - %s", lib_get_error( cd->lib_data ) );
	    return SMFIS_ACCEPT;
	    }
	}
    else if ( connaddr->sa_family == AF_INET6 )
	{
	struct sockaddr_in6* sa_in6;

	sa_in6 = (struct sockaddr_in6*) ( (void*) connaddr );	/* extra cast to avoid alignment warning */
	if ( whitelist != (iptab) 0 )
	    {
	    unsigned char* uchar_addr = (unsigned char*) &sa_in6->sin6_addr.s6_addr;
	    ipaddress ipa;
	    char str[100];

	    ipa.octets[0] = uchar_addr[0];
	    ipa.octets[1] = uchar_addr[1];
	    ipa.octets[2] = uchar_addr[2];
	    ipa.octets[3] = uchar_addr[3];
	    ipa.octets[4] = uchar_addr[4];
	    ipa.octets[5] = uchar_addr[5];
	    ipa.octets[6] = uchar_addr[6];
	    ipa.octets[7] = uchar_addr[7];
	    ipa.octets[8] = uchar_addr[8];
	    ipa.octets[9] = uchar_addr[9];
	    ipa.octets[10] = uchar_addr[10];
	    ipa.octets[11] = uchar_addr[11];
	    ipa.octets[12] = uchar_addr[12];
	    ipa.octets[13] = uchar_addr[13];
	    ipa.octets[14] = uchar_addr[14];
	    ipa.octets[15] = uchar_addr[15];
	    ipa.prefixlen = 128;
	    if ( iptab_check( whitelist, &ipa ) )
		{
		syslog( LOG_INFO, "whitelist \"%s\" [%s]", connhost, iptab_format_address( &ipa, str, sizeof(str) ) );
		cd->whitelisted = 1;
		}
	    }
	r = inet_ntop( AF_INET6, &sa_in6->sin6_addr, cd->buf1, sizeof(cd->buf1) );
	if ( r == (char*) 0 )
	    {
	    syslog( LOG_ERR, "couldn't convert IPv6 address to a string - %m" );
	    return SMFIS_ACCEPT;
	    }
	if ( ! lib_set_ipv6( cd->lib_data, sa_in6->sin6_addr, cd->buf1 ) )
	    {
	    syslog( LOG_ERR, "lib_set_ipv6() failed - %s", lib_get_error( cd->lib_data ) );
	    return SMFIS_ACCEPT;
	    }
	}
    else
	{
	syslog( LOG_ERR, "unknown address family - accepting" );
	return SMFIS_ACCEPT;
	}
    cd->ip_str = strdup( cd->buf1 );
    if ( cd->ip_str == (char*) 0 )
	{
	syslog( LOG_ERR, "couldn't strdup ip_str" );
	return SMFIS_ACCEPT;
	}

    return SMFIS_CONTINUE;
    }


/* spf_helo - handle the HELO command
**
** Called at the start of a connection.
**
** helohost: The string passed to the HELO/EHLO command.
*/
static sfsistat
spf_helo( SMFICTX* ctx, char* helohost )
    {
    connection_data_t* cd;

    cd = (connection_data_t*) smfi_getpriv( ctx );

    if ( cd->helo != (char*) 0 )
	{
	if ( strcmp( cd->helo, helohost ) != 0 )
	    syslog( LOG_NOTICE, "multiple HELOs from %s [%s] - ignoring '%s'", cd->helo, cd->ip_str, helohost );
	}
    else
	{
	cd->helo = strdup( helohost );
	if ( cd->helo == (char*) 0 )
	    {
	    syslog( LOG_ERR, "couldn't strdup helo" );
	    return SMFIS_ACCEPT;
	    }
	}

    if ( cd->whitelisted )
	return SMFIS_CONTINUE;

    if ( ! lib_set_helo_hostname( cd->lib_data, helohost ) )
	{
	syslog( LOG_ERR, "lib_set_helo_hostname() failed - %s", lib_get_error( cd->lib_data ) );
	return SMFIS_ACCEPT;
	}

    return SMFIS_CONTINUE;
    }


/* spf_envfrom - handle the MAIL FROM:<> command
**
** Called at the start of each message.  There may be multiple messages
** in a connection.  Any per-message data should be initialized here.
**
** fromargs: Null-terminated SMTP command arguments; fromargs[0] is the
**		   sender address.
*/
static sfsistat
spf_envfrom( SMFICTX* ctx, char** fromargs )
    {
    connection_data_t* cd;
    char* auth_type;
    char* verify;

    cd = (connection_data_t*) smfi_getpriv( ctx );

    if ( ! init_message_data( cd ) )
	{
	syslog( LOG_ERR, "init_message_data() failed" );
	return SMFIS_ACCEPT;
	}

    cd->mail_from = strdup( trim_address( fromargs[0] ) );
    if ( cd->mail_from == (char*) 0 )
	{
	syslog( LOG_ERR, "couldn't strdup mail_from" );
	return SMFIS_ACCEPT;
	}

    if ( cd->whitelisted )
	return SMFIS_CONTINUE;

    /* Authenticated messages get accepted without question. */
    auth_type = smfi_getsymval( ctx, "{auth_type}" );
    verify = smfi_getsymval( ctx, "{verify}" );
    if ( auth_type != (char*) 0 ||
	 ( verify != (char*) 0 && strcmp( verify, "OK" ) == 0 ) )
	{
	cd->authenticated = 1;
	return SMFIS_CONTINUE;
	}

    /* If the mail-from is empty or doesn't have an @, we make a fake one
    ** using the HELO domain (or if that's missing, the IP address).
    ** It's possible the SPF library we're using already does this, but at
    ** least some versions of some of the libraries do not, so we might
    ** as well do it here.
    */
    if ( cd->mail_from[0] == '\0' || strchr( cd->mail_from, '@' ) == (char*) 0 )
	{
	if ( cd->helo != (char*) 0 )
	    (void) snprintf( cd->buf1, sizeof(cd->buf1), "mailer_daemon@%s", cd->helo );
	else
	    (void) snprintf( cd->buf1, sizeof(cd->buf1), "mailer_daemon@[%s]", cd->ip_str );
	cd->use_mail_from = strdup( cd->buf1 );
	}
    else
	cd->use_mail_from = strdup( cd->mail_from );
    if ( cd->use_mail_from == (char*) 0 )
	{
	syslog( LOG_ERR, "couldn't strdup use_mail_from" );
	return SMFIS_ACCEPT;
	}

    if ( ! lib_set_from( cd->lib_data, cd->use_mail_from ) )
	{
	syslog( LOG_ERR, "lib_set_from() failed - %s", lib_get_error( cd->lib_data ) );
	return SMFIS_ACCEPT;
	}

    /* In MX mode we do a check for each recipient, and only later do the
    ** check for the whole message.
    */
    if ( recipientmx )
	return SMFIS_CONTINUE;

    if ( ! lib_do_check( cd->lib_data, cd->use_mail_from ) )
	{
	syslog( LOG_ERR, "lib_do_check() failed on mail from '%s' sent by [%s] - %s" , cd->use_mail_from, cd->ip_str, lib_get_error( cd->lib_data ) );
	return SMFIS_ACCEPT;
	}
    cd->result = lib_get_result( cd->lib_data );

    return handle_result( ctx );
    }


/* spf_envrcpt - handle a RCPT TO:<> command
**
** Called separately for each recipient of a message.
**
** rcptargs: Null-terminated SMTP command arguments; rcptargs[0] is the
**           recipient address.
*/
static sfsistat
spf_envrcpt( SMFICTX* ctx, char** rcptargs )
    {
    connection_data_t* cd;
    char* rcpt;

    cd = (connection_data_t*) smfi_getpriv( ctx );

    if ( cd->whitelisted || cd->authenticated )
	return SMFIS_CONTINUE;

    /* If we're not in MX mode then we've already checked the message. */
    if ( ! recipientmx )
	return SMFIS_CONTINUE;

    rcpt = trim_address( rcptargs[0] );

    if ( ! lib_do_check_recipient( cd->lib_data, rcpt ) )
	{
	syslog( LOG_ERR, "lib_do_check_recipient() failed - %s", lib_get_error( cd->lib_data ) );
	return SMFIS_ACCEPT;
	}
    cd->result = lib_get_result( cd->lib_data );

    return handle_result( ctx );
    }


/* spf_header - handle a header line
**
** Called separately for each header line in a message.
**
** name:  Header field name.
** value: Header vield value, including folded whitespace.  The final CRLF
**		is removed.
*/
static sfsistat
spf_header( SMFICTX* ctx, char* name, char* value )
    {
    connection_data_t* cd;
    
    cd = (connection_data_t*) smfi_getpriv( ctx );

    /* We look for any instances of our own header in the incoming
    ** message, so we can remove them later.
    */
    if ( strcasecmp( name, HEADER_NAME ) == 0 )
	{
	++cd->n_my_headers;

	/* Does this SPF header claim to be from us? */
	if ( receiver_is_me( value ) )
	    {
	    syslog( LOG_NOTICE, "found possible spoofed header - '%s'", value );
	    if ( cd->n_del_headers < DEL_HEADER_MAX )
		{
		cd->del_header_list[cd->n_del_headers] = cd->n_my_headers;
		++cd->n_del_headers;
		}
	    }
	}

    return SMFIS_CONTINUE;
    }


/* spf_eoh - handle the end of the headers
**
** Called once per message after all headers have been processed.
*/
static sfsistat
spf_eoh( SMFICTX* ctx )
    {
    connection_data_t* cd;
    
    cd = (connection_data_t*) smfi_getpriv( ctx );

    if ( cd->whitelisted || cd->authenticated )
	return SMFIS_CONTINUE;

    if ( ! recipientmx )
	return SMFIS_CONTINUE;

    if ( ! lib_do_check_final( cd->lib_data ) )
	{
	syslog( LOG_ERR, "lib_do_check_final() failed - %s", lib_get_error( cd->lib_data ) );
	return SMFIS_ACCEPT;
	}
    cd->result = lib_get_result( cd->lib_data );

    cd->action = SPFMILTER_ACTION_MARK;
    return SMFIS_CONTINUE;
    }


/* spf_eom - handle the end of the message
**
** Called once per message after all body blocks have been processed.
** Any per-message data should be freed both here and in spf_abort(),
** and wherever you return any of SMFIS_REJECT/SMFIS_TEMPFAIL/SMFIS_DISCARD.
*/
static sfsistat
spf_eom( SMFICTX* ctx )
    {
    connection_data_t* cd;
    int i;

    cd = (connection_data_t*) smfi_getpriv( ctx );

    /* Header deleting and adding can only happen in the eom handler. */

    /* Remove any previous SPF headers, to prevent spoofing. */
    if ( cd->n_del_headers >= DEL_HEADER_MAX )
	{
	/* There were too many bogus SPF headers to keep track of, so to
	** be safe we just delete all SPF headers.
	*/
	for ( i = cd->n_my_headers; i >= 1; --i )
	    smfi_chgheader( ctx, HEADER_NAME, i, (char*) 0 );
	}
    else
	{
	/* We have a list of only the headers purporting to be from us;
	** those get deleted.  We do the deletions in reverse order in case
	** deleting one header causes subsequent headers to get re-numbered.
	** (I actually checked whether this happens in the current milter
	** implementation, and it does NOT, but perhaps some future version
	** will get it wrong.)
	*/
	for ( i = cd->n_del_headers - 1 ; i >= 0; --i )
	    smfi_chgheader( ctx, HEADER_NAME, cd->del_header_list[i], (char*) 0 );
	}

    if ( cd->whitelisted || cd->authenticated )
	{
	cd->result = SPFMILTER_RESULT_PASS;
	cd->action = SPFMILTER_ACTION_MARK;
	if ( cd->whitelisted )
	    (void) snprintf( cd->buf1, sizeof(cd->buf1), "%s: %s is whitelisted", local_hostname, cd->ip_str );
	else if ( cd->authenticated )
	    (void) snprintf( cd->buf1, sizeof(cd->buf1), "%s: authenticated connection", local_hostname );
	else
	    (void) strncpy( cd->buf1, "???", sizeof(cd->buf1) );
	}
    else
	reason_str( cd, cd->buf1, sizeof(cd->buf1) );

    if ( cd->action == SPFMILTER_ACTION_MARK )
	{
	/* Build and add the SPF header. */
	build_header( cd, cd->buf2, sizeof(cd->buf2), cd->buf1 );
	/* The SPF spec wants us to prepend the new header to the existing
	** ones, but not all versions of the milter API have the necessary
	** function.
	*/
#ifdef HAVE_SMFI_INSHEADER
	smfi_insheader( ctx, 0, HEADER_NAME, cd->buf2 );
#else
	smfi_addheader( ctx, HEADER_NAME, cd->buf2 );
#endif
	}

    term_message_data( cd );

    return SMFIS_CONTINUE;
    }


/* spf_abort - handle the message being aborted
**
** May be called at any time during processing of a message, indicating
** that something went wrong.  Any per-message data should be freed both
** here and in sample_eom(), and wherever you return any of
** SMFIS_REJECT/SMFIS_TEMPFAIL/SMFIS_DISCARD.
*/
static sfsistat
spf_abort( SMFICTX* ctx )
    {
    connection_data_t* cd;
    
    cd = (connection_data_t*) smfi_getpriv( ctx );

    if ( cd != (connection_data_t*) 0 )
	term_message_data( cd );
    else
	syslog( LOG_NOTICE, "spf_abort called but cd is null" );

    return SMFIS_CONTINUE;
    }


/* spf_close - handle the connection being closed
**
** Called once at the end of a connection.  Any per-connection data
** should be freed here.
*/
static sfsistat
spf_close( SMFICTX* ctx )
    {
    connection_data_t* cd;

    cd = (connection_data_t*) smfi_getpriv( ctx );

    if ( cd != (connection_data_t*) 0 )
	{
	smfi_setpriv( ctx, (void*) 0 );
	term_connection_data( cd );
	}
    else
	syslog( LOG_NOTICE, "spf_close called but cd is null" );

    return SMFIS_CONTINUE;
    }


static char*
trim_address( char* address )
    {
    int len;

    /* The milter API sometimes gives us strings with angle brackets. */
    while ( *address == '<' )
	++address;
    len = strlen( address );
    while ( len > 0 && address[len - 1] == '>' )
	address[--len] = '\0';

    /* We also sometimes get CRs on the ends of addresses. */
    while ( len > 0 && address[len - 1] == '\015' )
	address[--len] = '\0';

    return address;
    }


static sfsistat
handle_result( SMFICTX* ctx )
    {
    connection_data_t* cd;
    const char* exp;
    char exp_escaped[1000];

    cd = (connection_data_t*) smfi_getpriv( ctx );

    switch ( cd->result )
	{
	case SPFMILTER_RESULT_PASS:
	case SPFMILTER_RESULT_SOFTFAIL:
	case SPFMILTER_RESULT_NEUTRAL:
	case SPFMILTER_RESULT_UNKNOWN:
	case SPFMILTER_RESULT_UNMECH:
	case SPFMILTER_RESULT_INVALID:
	case SPFMILTER_RESULT_TEMPERROR:
	case SPFMILTER_RESULT_ERROR:
	case SPFMILTER_RESULT_NONE:
	    cd->action = SPFMILTER_ACTION_MARK;
	    break;
	case SPFMILTER_RESULT_FAIL:
	    if ( markonly )
		cd->action = SPFMILTER_ACTION_MARK;
	    else
		cd->action = SPFMILTER_ACTION_REJECT;
	    break;
	}

    switch ( cd->action )
	{
	case SPFMILTER_ACTION_REJECT:
	    exp = lib_get_explanation( cd->lib_data );
	    syslog( LOG_INFO, "rejecting mail from [%s] - %s", cd->ip_str, result_str( cd->result ) );
	    if ( exp != (char*) 0 )
		escape_percents( exp, exp_escaped, sizeof(exp_escaped) );
	    else
		(void) strncpy( exp_escaped, "rejected by spfmilter", sizeof(exp_escaped) - 1 );
	    smfi_setreply( ctx, "550", "5.7.1", exp_escaped );
	    term_message_data( cd );
	    return SMFIS_REJECT;

	case SPFMILTER_ACTION_TEMPFAIL:
	    syslog( LOG_INFO, "temporarily failing mail from [%s] - %s", cd->ip_str, lib_get_error( cd->lib_data ) );
	    smfi_setreply( ctx, "421", "4.3.2", "Please try again later." );
	    term_message_data( cd );
	    return SMFIS_TEMPFAIL;

	default:
	    return SMFIS_CONTINUE;
	}
    }


/* escape_percents - escape '%' characters in a string
**
** Sendmail doesn't like sending percent characters in bounce messages.
** If a milter returns an error message with percents in it, sendmail
** will ignore that and instead return an uninformative default
** message saying "Command rejected".  To avoid this we escape the
** percents witrh a second percent.
*/
static void
escape_percents( const char* in, char* out, int out_size )
    {
    const char* i;
    char* o;
    int out_len;

    i = in;
    o = out;
    out_len = 0;
    while ( *i != '\0' && out_len + 2 < out_size )
	{
	if ( *i == '%' )
	    {
	    /* The character that most often comes back escaped is
	    ** the @ sign (%40). Since it doesn't actually need
	    ** escaping, unescape it.
	    */
	    if ( i[1] == '4' && i[2] == '0' )
		{
		*o++ = '@';
		++out_len;
		i += 3;
		}
	    else
		{
		/* Otherwise, escape the %-sign. */
		*o++ = '%';
		*o++ = '%';
		out_len += 2;
		++i;
		}
	    }
	else
	    {
	    *o++ = *i++;
	    ++out_len;
	    }
	}
    *o = '\0';
    }


static int
receiver_is_me( const char* value )
    {
    const char* cp;

    /* First look for a receiver= setting after the comment part.
    ** The hostname immediately follows the equals and is immediately
    ** followed by a semicolon.
    */
    cp = strrchr( value, ')' );
    if ( cp != (char*) 0 )
	++cp;
    else
	cp = value;
    cp = strstr( cp, " receiver=" );
    if ( cp != (char*) 0 )
	{
	cp += 10;
	if ( strncasecmp( cp, local_hostname, local_hostname_len ) == 0 )
	    {
	    cp += local_hostname_len;
	    if ( *cp == ';' )
		return 1;	/* Yes! */
	    }
	}

    /* If there was no receiver= (it's optional), try checking the
    ** comment part for a hostname.  That's even more optional
    ** (it's not in the spec at all), but many implementations do add it.
    **
    ** To find the comment hostname, skip one word, whitespace, and
    ** an open-paren, then check for the hostname followed by a colon.
    */
    cp = value;
    cp += strcspn( cp, " \t" );
    cp += strspn( cp, " \t" );
    if ( *cp == '(' )
	{
	++cp;
	if ( strncasecmp( cp, local_hostname, local_hostname_len ) == 0 )
	    {
	    cp += local_hostname_len;
	    if ( *cp == ':' )
		return 1;	/* Yes! */
	    }
	}

    /* Nope. */
    return 0;
    }


static void
build_header( connection_data_t* cd, char* header, size_t header_size, const char* reason )
    {
    int len;

    (void) snprintf( header, header_size, "%s", result_str( cd->result ) );
    len = strlen( header );
    if ( reason != (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " (%s)", reason );
	len = strlen( header );
	}
    if ( local_hostname != (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " receiver=%s;", local_hostname );
	len = strlen( header );
	}
    if ( cd->ip_str!= (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " client-ip=%s;", cd->ip_str);
	len = strlen( header );
	}
    if ( cd->helo != (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " helo=%s;", cd->helo );
	len = strlen( header );
	}
    if ( cd->mail_from != (char*) 0 )
	{
	(void) snprintf( &header[len], header_size - len, " envelope-from=%s;", cd->mail_from );
	len = strlen( header );
	}
    /*!!! Do something about the problem= field. */
    (void) snprintf( &header[len], header_size - len, " x-software=%s %s %s with %s;", SPFMILTER_PROGRAM, SPFMILTER_VERSION, SPFMILTER_URL, lib_version() );
    }


static connection_data_t*
init_connection_data( void )
    {
    connection_data_t* cd;

    cd = (connection_data_t*) malloc( sizeof(connection_data_t) );
    if ( cd == (connection_data_t*) 0 )
	return (connection_data_t*) 0;
    cd->whitelisted = 0;
    cd->ip_str = (char*) 0;
    cd->helo = (char*) 0;
    cd->lib_data = lib_init_connection_data();
    if ( cd->lib_data == (lib_data_t*) 0 )
	{
	syslog( LOG_ERR, "lib_init_connection_data() failed" );
	free( (void*) cd );
	return (connection_data_t*) 0;
	}
    if ( ! init_message_data( cd ) )
	{
	syslog( LOG_ERR, "init_message_data() failed" );
	term_connection_data( cd );
	return (connection_data_t*) 0;
	}
    return cd;
    }


static int
init_message_data( connection_data_t* cd )
    {
    cd->authenticated = 0;
    cd->mail_from = (char*) 0;
    cd->use_mail_from = (char*) 0;
    cd->result = -1;
    cd->action = SPFMILTER_ACTION_UNKNOWN;
    cd->n_my_headers = 0;
    cd->n_del_headers = 0;
    if ( ! lib_init_message_data( cd->lib_data ) )
	{
	syslog( LOG_ERR, "lib_init_message_data() failed" );
	return 0;
	}
    return 1;
    }


/* Frees or resets any items in the connection_data that are for
** an individual message.
*/
static void
term_message_data( connection_data_t* cd )
    {
    if ( cd->mail_from != (char*) 0 )
	{
	free( (void*) cd->mail_from );
	cd->mail_from = (char*) 0;
	}
    if ( cd->use_mail_from != (char*) 0 )
	{
	free( (void*) cd->use_mail_from );
	cd->use_mail_from = (char*) 0;
	}
    lib_term_message_data( cd->lib_data );
    }


/* Frees all items in the connection_data. */
static void
term_connection_data( connection_data_t* cd )
    {
    term_message_data( cd );
    if ( cd->ip_str != (char*) 0 )
	free( (void*) cd->ip_str );
    if ( cd->helo != (char*) 0 )
	free( (void*) cd->helo );
    lib_term_connection_data( cd->lib_data );
    free( (void*) cd );
    }


static char*
result_str( int result )
    {
    switch ( result )
	{
	case SPFMILTER_RESULT_PASS: return "pass";
	case SPFMILTER_RESULT_FAIL: return "fail";
	case SPFMILTER_RESULT_SOFTFAIL: return "softfail";
	case SPFMILTER_RESULT_NEUTRAL: return "neutral";
	case SPFMILTER_RESULT_UNKNOWN: return "unknown";
	case SPFMILTER_RESULT_UNMECH: return "unknown-mechanisms";
	case SPFMILTER_RESULT_INVALID: return "invalid";
	case SPFMILTER_RESULT_TEMPERROR: return "temporary-error";
	case SPFMILTER_RESULT_ERROR: return "error";
	case SPFMILTER_RESULT_NONE: return "none";
	default: return "???";
	}
    }


static void
reason_str( connection_data_t* cd, char* str, size_t size )
    {
    switch ( cd->result )
	{
	case SPFMILTER_RESULT_PASS:
	(void) snprintf( str, size, "%s: domain of %s designates %s as permitted sender", local_hostname, cd->use_mail_from, cd->ip_str );
	break;
	case SPFMILTER_RESULT_FAIL:
	(void) snprintf( str, size, "%s: domain of %s does not designate %s as permitted sender", local_hostname, cd->use_mail_from, cd->ip_str );
	break;
	case SPFMILTER_RESULT_SOFTFAIL:
	(void) snprintf( str, size, "%s: domain of transitioning %s does not designate %s as permitted sender", local_hostname, cd->use_mail_from, cd->ip_str );
	break;
	case SPFMILTER_RESULT_NEUTRAL:
	(void) snprintf( str, size, "%s: %s is neither permitted nor denied by domain of %s", local_hostname, cd->ip_str, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_UNKNOWN:
	(void) snprintf( str, size, "%s: unknown result during lookup of %s", local_hostname, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_UNMECH:
	(void) snprintf( str, size, "%s: domain of %s uses a mechanism not recognized by this client", local_hostname, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_INVALID:
	(void) snprintf( str, size, "%s: domain of %s has invalid SPF", local_hostname, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_TEMPERROR:
	(void) snprintf( str, size, "%s: temporary error in processing during lookup of %s", local_hostname, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_ERROR:
	(void) snprintf( str, size, "%s: error in processing during lookup of %s", local_hostname, cd->use_mail_from );
	break;
	case SPFMILTER_RESULT_NONE:
	(void) snprintf( str, size, "%s: %s does not designate permitted sender hosts", local_hostname, cd->use_mail_from );
	break;
	default:
	(void) snprintf( str, size, "%s: ???", local_hostname );
	break;
	}
    }


/* Decide which library we can/should use. */
#ifdef HAVE_SPF_H
#define USE_LIBSPF
#endif
#ifdef HAVE_SPF2_SPF_H
#define USE_LIBSPF2
#endif

#if !defined(USE_LIBSPF) && !defined(USE_LIBSPF2)
#error "neither libspf nor libspf2 is present - please provide one"
#endif

#if defined(USE_LIBSPF) && defined(USE_LIBSPF2)
#undef USE_LIBSPF	/* if both are present, use libspf2 */
#endif


#ifdef USE_LIBSPF

/* Libspf data and routines. */


#include <spf.h>


struct lib_fallback_s {
    int junk;	/* not used, but mallocing zero bytes can be a problem */
    };


struct lib_data_s {
    peer_info_t* peer_info;
    SPF_RESULT res;
    };


static int
lib_init( void )
    {
    if ( recipientmx )
	{
	syslog( LOG_ERR, "sorry, --recipientmx mode is not implemented with libspf; if you want to use that flag you will need to rebuild spfmilter with libspf2" );
	return 0;
	}

    /* SPF_dbg_level( debug ); */

    return 1;
    }


static const char*
lib_version( void )
    {
    return "libspf-unknown";	/* the library doesn't tell us its version */
    }


static lib_fallback_t*
lib_init_fallback( const char* str )
    {
    lib_fallback_t* lf;

    lf = (lib_fallback_t*) malloc( sizeof(lib_fallback_t) );
    if ( lf == (lib_fallback_t*) 0 )
	return (lib_fallback_t*) 0;
    return lf;
    }


static int
lib_set_local_hostname( lib_data_t* ld )
    {
    return 1;
    }


static lib_data_t*
lib_init_connection_data( void )
    {
    lib_data_t* ld;

    ld = (lib_data_t*) malloc( sizeof(lib_data_t) );
    if ( ld == (lib_data_t*) 0 )
	return (lib_data_t*) 0;
    return ld;
    }


static int
lib_init_message_data( lib_data_t* ld )
    {
    return 1;
    }


static int
lib_set_ipv4( lib_data_t* ld, struct in_addr ipv4_addr, char* ipv4_str )
    {
    ld->peer_info = SPF_init( local_hostname, ipv4_str, explanation_str, (char*) 0, (char*) 0, trustedforwarders, 0 );
    if ( ld->peer_info == (peer_info_t*) 0 )
	return 0;
    ld->res = -1;
    return 1;
    }


static int
lib_set_ipv6( lib_data_t* ld, struct in6_addr ipv6_addr, char* ipv6_str )
    {
    ld->peer_info = SPF_init( local_hostname, ipv6_str, explanation_str, (char*) 0, (char*) 0, trustedforwarders, 0 );
    if ( ld->peer_info == (peer_info_t*) 0 )
	return 0;
#ifdef notdef
    ld->res = -1;
    return 1;
#else /* notdef */
    syslog( LOG_ERR, "sorry, libspf doesn't actually wortk with IPv6; if you want to use v6 you will need to rebuild spfmilter with libspf2" );
    return 0;
#endif /* notdef */
    }


static int
lib_set_helo_hostname( lib_data_t* ld, char* helo_hostname )
    {
    if ( ! SPF_smtp_helo( ld->peer_info, helo_hostname ) )
	return 0;
    return 1;
    }


static int
lib_set_from( lib_data_t* ld, const char* from )
    {
    if ( ! SPF_smtp_from( ld->peer_info, from ) )
	return 0;
    return 1;
    }


static int
find_fallback( const char* from )
    {
    const char* from_host;
    int i;

    from_host = strchr( from, '@' );
    if ( from_host == (char*) 0 )
	return -1;
    ++from_host;
    if ( fallbacks == (fallback_t*) 0 )
	return -1;
    for ( i = 0; i < num_fallbacks; ++i )
	if ( match( fallbacks[i].pattern, from_host ) )
	    return i;
    return -1;
    }


static int
lib_do_check( lib_data_t* ld, const char* from )
    {
    ld->res = SPF_policy_main( ld->peer_info );

    /* If there was no SPF record, then try finding a fallback mechanism. */
    if ( ld->res == SPF_NONE )
	{
	int i = find_fallback( from );

	if ( i != -1 )
	    {
	    if ( ! SPF_parse_policy( ld->peer_info, fallbacks[i].str ) )
		return 0;
	    ld->res = ld->peer_info->RES;
	    }
	}

    return 1;
    }


static int
lib_do_check_recipient( lib_data_t* ld, const char* to )
    {
    /* Not implemented for libspf. */
    return 0;
    }


static int
lib_do_check_final( lib_data_t* ld )
    {
    /* Not implemented for libspf. */
    return 0;
    }


static int
lib_get_result( lib_data_t* ld )
    {
    /* Convert libspf result to spfmilter result. */
    switch ( ld->res )
	{
	case SPF_PASS: return SPFMILTER_RESULT_PASS;
	case SPF_NONE: return SPFMILTER_RESULT_NONE;
	case SPF_S_FAIL: return SPFMILTER_RESULT_SOFTFAIL;
	case SPF_H_FAIL: return SPFMILTER_RESULT_FAIL;
	case SPF_ERROR: return SPFMILTER_RESULT_ERROR;
	case SPF_NEUTRAL: return SPFMILTER_RESULT_NEUTRAL;
	case SPF_UNKNOWN: return SPFMILTER_RESULT_UNKNOWN;
	case SPF_UNMECH: return SPFMILTER_RESULT_UNMECH;
	default: return -1;
	}
    }


static const char*
lib_get_explanation( lib_data_t* ld )
    {
    return ld->peer_info->explain;
    }


static const char*
lib_get_error( lib_data_t* ld )
    {
    return ld->peer_info->error;
    }


static void
lib_term_message_data( lib_data_t* ld )
    {
    }


static void
lib_term_connection_data( lib_data_t* ld )
    {
    if ( ld != (lib_data_t*) 0 )
	{
	if ( ld->peer_info != (peer_info_t*) 0 )
	    ld->peer_info = SPF_close( ld->peer_info );
	free( (void*) ld );
	}
    }


static void
lib_term_fallback( lib_fallback_t* lf )
    {
    if ( lf != (lib_fallback_t*) 0 )
	free( (void*) lf );
    }


static void
lib_term( void )
    {
    }

#endif /* USE_LIBSPF */


#ifdef USE_LIBSPF2

/* Libspf2 data and routines. */

#include <spf2/spf.h>

static SPF_server_t* spf_server;

struct lib_fallback_s {
    /*!!!*/
    int junk;
    };

struct lib_data_s {
    SPF_request_t* spf_request;
    SPF_response_t* spf_response;
    SPF_result_t spf_result;
    };


static int
lib_init( void )
    {
    spf_server = SPF_server_new( SPF_DNS_CACHE, debug );
    if ( spf_server == (SPF_server_t*) 0 )
	{
	syslog( LOG_ERR, "SPF_server_new() failed" );
	return 0;
	}

    if ( localpolicy_str != (char*) 0 )
	{
	int err, i;
	SPF_response_t* lp_response = (SPF_response_t*) 0;

	err = SPF_server_set_localpolicy( spf_server, localpolicy_str, trustedforwarders, &lp_response );
	if ( err )
	    {
	    syslog( LOG_ERR, "error compiling local policy - %d (%s)", err, SPF_strerror( err ) );
	    for ( i = 0; i < SPF_response_messages( lp_response ); ++i )
		{
		SPF_error_t* lp_error = SPF_response_message( lp_response, i );
		syslog( LOG_ERR, "  %s: %s", 
		    SPF_error_errorp( lp_error ) ? "Error" : "Warning",
		    SPF_error_message( lp_error ) );
		}
	    if ( lp_response != (SPF_response_t*) 0 )
		SPF_response_free( lp_response );
	    return 0;
	    }
	if ( lp_response != (SPF_response_t*) 0 )
	    SPF_response_free( lp_response );
	}

    if ( explanation_str != (char*) 0 )
	{
	int err, i;
	SPF_response_t* ex_response;

	err = SPF_server_set_explanation( spf_server, explanation_str, &ex_response );
	if ( err )
	    {
	    syslog( LOG_ERR, "error compiling explanation - %d (%s)", err, SPF_strerror( err ) );
	    for ( i = 0; i < SPF_response_messages( ex_response ); ++i )
		{
		SPF_error_t* ex_error = SPF_response_message( ex_response, i );
		syslog( LOG_ERR, "  %s: %s",
		    SPF_error_errorp( ex_error ) ? "Error" : "Warning",
		    SPF_error_message( ex_error ) );
		}
	    SPF_response_free( ex_response );
	    return 0;
	    }
	SPF_response_free( ex_response );
	}

    return 1;
    }


static const char*
lib_version( void )
    {
    static char version[100];

    (void) snprintf( version, sizeof(version), "libspf2-%d.%d.%d", SPF_LIB_VERSION_MAJOR, SPF_LIB_VERSION_MINOR, SPF_LIB_VERSION_PATCH );
    return version;
    }


static lib_fallback_t*
lib_init_fallback( const char* str )
    {
    lib_fallback_t* lf;

    lf = (lib_fallback_t*) malloc( sizeof(lib_fallback_t) );
    if ( lf == (lib_fallback_t*) 0 )
	return (lib_fallback_t*) 0;
    /*!!!*/
    return lf;
    }


static int
lib_set_local_hostname( lib_data_t* ld )
    {
    if ( SPF_server_set_rec_dom( spf_server, local_hostname ) )
	{
	syslog( LOG_ERR, "SPF_server_set_rec_dom() failed" );
	return 0;
	}
    return 1;
    }


static lib_data_t*
lib_init_connection_data( void )
    {
    lib_data_t* ld;

    ld = (lib_data_t*) malloc( sizeof(lib_data_t) );
    if ( ld == (lib_data_t*) 0 )
	return (lib_data_t*) 0;

    ld->spf_request = SPF_request_new( spf_server );
    if ( ld->spf_request == (SPF_request_t*) 0 )
	{
	syslog( LOG_ERR, "SPF_request_new() failed" );
	free( (void*) ld );
	return (lib_data_t*) 0;
	}
    ld->spf_response = (SPF_response_t*) 0;

    return ld;
    }


static int
lib_init_message_data( lib_data_t* ld )
    {
    return 1;
    }


static int
lib_set_ipv4( lib_data_t* ld, struct in_addr ipv4_addr, char* ipv4_str )
    {
    if ( SPF_request_set_ipv4_str( ld->spf_request, ipv4_str ) )
	{
	syslog( LOG_ERR, "SPF_request_set_ipv4_str() failed" );
	return 0;
	}
    return 1;
    }


static int
lib_set_ipv6( lib_data_t* ld, struct in6_addr ipv6_addr, char* ipv6_str )
    {
    if ( SPF_request_set_ipv6_str( ld->spf_request, ipv6_str ) )
	{
	syslog( LOG_ERR, "SPF_request_set_ipv6_str() failed" );
	return 0;
	}
    return 1;
    }


static int
lib_set_helo_hostname( lib_data_t* ld, char* helo_hostname )
    {
    if ( SPF_request_set_helo_dom( ld->spf_request, helo_hostname ) )
	{
	syslog( LOG_ERR, "SPF_request_set_helo_dom() failed" );
	return 0;
	}
    return 1;
    }


static int
lib_set_from( lib_data_t* ld, const char* from )
    {
    if ( SPF_request_set_env_from( ld->spf_request, from ) )
	{
	syslog( LOG_ERR, "SPF_request_set_env_from() failed" );
	return 0;
	}
    return 1;
    }


static int
lib_do_check( lib_data_t* ld, const char* from )
    {
    ld->spf_response = (SPF_response_t*) 0;
    if ( SPF_request_query_mailfrom( ld->spf_request, &ld->spf_response ) )
	{
	syslog( LOG_ERR, "SPF_request_query_mailfrom() failed" );
	return 0;
	}
    ld->spf_result = SPF_response_result( ld->spf_response );
    return 1;
    }


static int
lib_do_check_recipient( lib_data_t* ld, const char* to )
    {
    /*!!!*/
    return 1;
    }


static int
lib_do_check_final( lib_data_t* ld )
    {
    /*!!!*/
    return 1;
    }


static int
lib_get_result( lib_data_t* ld )
    {
    /* Convert libspf2 result to spfmilter result. */
    switch ( ld->spf_result )
	{
	case SPF_RESULT_INVALID: return SPFMILTER_RESULT_INVALID;
	case SPF_RESULT_NEUTRAL: return SPFMILTER_RESULT_NEUTRAL;
	case SPF_RESULT_PASS: return SPFMILTER_RESULT_PASS;
	case SPF_RESULT_FAIL: return SPFMILTER_RESULT_FAIL;
	case SPF_RESULT_SOFTFAIL: return SPFMILTER_RESULT_SOFTFAIL;
	case SPF_RESULT_NONE: return SPFMILTER_RESULT_NONE;
	case SPF_RESULT_TEMPERROR: return SPFMILTER_RESULT_TEMPERROR;
	case SPF_RESULT_PERMERROR: return SPFMILTER_RESULT_ERROR;
	default: return -1;
	}
    }


static const char*
lib_get_explanation( lib_data_t* ld )
    {
    return SPF_strresult( ld->spf_result );
    }


static const char*
lib_get_error( lib_data_t* ld )
    {
    return SPF_strerror( SPF_response_errcode( ld->spf_response ) );
    }


static void
lib_term_message_data( lib_data_t* ld )
    {
    }


static void
lib_term_connection_data( lib_data_t* ld )
    {
    if ( ld != (lib_data_t*) 0 )
	{
	if ( ld->spf_request != (SPF_request_t*) 0 )
	    SPF_request_free( ld->spf_request );
	if ( ld->spf_response != (SPF_response_t*) 0 )
	    SPF_response_free( ld->spf_response );
	free( (void*) ld );
	}
    }


static void
lib_term_fallback( lib_fallback_t* lf )
    {
    if ( lf != (lib_fallback_t*) 0 )
	{
	/*!!!*/
	free( (void*) lf );
	}
    }


static void
lib_term( void )
    {
    SPF_server_free( spf_server );
    }

#endif /* USE_LIBSPF2 */
